! ESA (C) 2000-2018
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module esp_smooth_mod

  use global

  implicit none

  private

  public :: esp_smooth

contains

!-------------------------------------------------------------------------------
! ESP_SMOOTH   ---  Iris Traulsen (AIP), adapted from Georg Lamer's erbackmap
  subroutine esp_smooth

    ! cheesed image:   iches
    ! cheesed mask:    mfeld
    ! exposure map:    iexmp
    ! background map:  iback

    !--------------------------------------------------------------------------
    ! I. Expand all input images by <iext> pixels at each side (1/4 in length)
    ! in order to avoid border / image wrapping effects with fft.
    ! Copy stripe of pixel data to expansion region.
    ! II. Smooth temporary background .
    ! III. Copy relevant part of resulting backgroud map to backmap.
    !-------------------------------------------------------------------------

    include 'include/cdbio_f90.cmn'

    real, dimension(:,:), allocatable :: image_ext
    real, dimension(:,:), allocatable :: expimage_ext
    real, dimension(:,:), allocatable :: backmap_ext
    real, dimension(:,:), allocatable :: cheesemask_ext

    integer         :: status
    integer         :: iext

    real           :: sigma,snrmin,snrdiff,snrup,snrlow
    real,parameter :: snreps=1E-2
    integer         :: nsig,isig,ix,iy,imin
    integer,dimension(1) :: iminarr

    integer,dimension(:,:),allocatable   :: mindex

    real,dimension(:,:,:), allocatable :: bkgarr
    real,dimension(:,:,:), allocatable :: snrarr

    real,dimension(:,:), allocatable :: kernel,image1,image2,image3
    real,dimension(:,:), allocatable :: smoothed,msk_smoothed,exp_smoothed
    real,dimension(:,:), allocatable :: snr,counts
    real                             :: knorm, climit, ctest

    real,dimension(ipa_x) :: isumx
    real,dimension(ipa_y) :: isumy
    integer               :: ixmin, ixmax, iymin, iymax, ixsize, iysize

    real  :: sqrt2=1.4142135624

    status=0

    iback = 0

    !! subimage of (1:ipa_x,1:ipa_y)                   !!! IT 2019/03
    !! code: cf emldetect, eboxdetect --> bdfi4r.f90

    !! internal values:
    ! width of extension strip, rounded to even number
    ! rectangular images: use the longer side / 4
    ! pn small-window mode: use the full length of the shorter side
    iext=min(max(max(iusex, iusey)/8*2+2, 200), iusex/2*2, iusey/2*2)
    ! increasing smoothing kernel: number of steps
    nsig=8
    ! maximum tolerated relative deviation between input and output count
    ! number before triggering a warning
    climit = 0.1

    if  (iusex < iext .or. iusey < iext) then
       call error('Mismatch', &
                  'Internal error: Expansion region > original image')
    endif

!!!    ixsize=iusex+2*iext
!!!    iysize=iusey+2*iext
!!! TEST: use a square
    ixsize = max(iusex+2*iext, iusey+2*iext)
    iysize = ixsize
    ixmin = ixsize/2-iusex/2
    iymin = iysize/2-iusey/2
    ixmax = ixmin + ix1-ix0
    iymax = iymin + iy1-iy0
    allocate(image_ext(ixsize,iysize))
    image_ext = 0.
    allocate(cheesemask_ext(ixsize,iysize))
    cheesemask_ext = 0.
    allocate(backmap_ext(ixsize,iysize))
    backmap_ext = 0.
    if (db_exp) then
       allocate(expimage_ext(ixsize,iysize))
       expimage_ext = 0.
    endif

!    cheesemask_ext(iext+1:ixsize-iext , iext+1:iysize-iext) &
    cheesemask_ext(ixmin:ixmax, iymin:iymax) = real(MFELD(ix0:ix1,iy0:iy1))
    cheesemask_ext(iext:1:-1 , 1:iysize) = &
         cheesemask_ext(iext+1:2*iext , 1:iysize)
    cheesemask_ext(ixsize:ixsize-iext+1:-1 , 1:iysize) = &
         cheesemask_ext(ixsize-2*iext+1:ixsize-iext , 1:iysize)
    cheesemask_ext(1:ixsize , iext:1:-1) = &
         cheesemask_ext(1:ixsize , iext+1:2*iext)
    cheesemask_ext(1:ixsize , iysize:iysize-iext+1:-1) = &
         cheesemask_ext(1:ixsize , iysize-2*iext+1:iysize-iext)
    ! if a non-standard mask with values > 1 is provided: reset them to 1
    ! (for multiplication later on)
    where (cheesemask_ext.gt.0)
       cheesemask_ext = 1
    endwhere

    ! 2021/04: mask bright OOT regions
    !          --  so far using fixed cuts (see global.f90)
    ! Cut was derived empirically for standard 4"x4" bins and is thus scaled
    ! to the actual pixel size using crdeltI here
    if (withoot) then
      if (idinst .eq. 1) then
        do iy=iymin,iymax
          do ix=ixmin,ixmax
            if (oot_sky(ix0+ix-ixmin,iy0+iy-iymin) &
                 .gt. ootcut*abs(crdelt1)*abs(crdelt2)*810000.) &
              cheesemask_ext( min(max(ix-ootpix,1),ixsize)    &
                             :min(max(ix+ootpix,1),ixsize),   &
                              min(max(iy-ootpix,1),iysize)    &
                             :min(max(iy+ootpix,1),iysize))  =  0.0
          end do
        end do
      end if
    end if

    ! mirror <iext> pixels wide stripe of image about its original border
!    image_ext(iext+1:ixsize-iext , iext+1:iysize-iext) &
    image_ext(ixmin:ixmax, iymin:iymax) = &
         ICHES(ix0:ix1,iy0:iy1) * cheesemask_ext(ixmin:ixmax, iymin:iymax)
    ! fill masked areas with small values
    where (cheesemask_ext.eq.0)
       image_ext = 0.5*sum(ICHES(ix0:ix1,iy0:iy1))/sum(MFELD(ix0:ix1,iy0:iy1))
    endwhere
    image_ext(iext:1:-1 , 1:iysize) = &
         image_ext(iext+1:2*iext , 1:iysize)
    image_ext(ixsize:ixsize-iext+1:-1 , 1:iysize) = &
         image_ext(ixsize-2*iext+1:ixsize-iext , 1:iysize)
    image_ext(1:ixsize , iext:1:-1) = &
         image_ext(1:ixsize , iext+1:2*iext)
    image_ext(1:ixsize , iysize:iysize-iext+1:-1) = &
         image_ext(1:ixsize , iysize-2*iext+1:iysize-iext)

    if (db_exp) then
!    expimage_ext(iext+1:ixsize-iext , iext+1:iysize-iext) &
    expimage_ext(ixmin:ixmax, iymin:iymax) = IEXMP(ix0:ix1,iy0:iy1)
    expimage_ext(iext:1:-1 , 1:iysize) = &
         expimage_ext(iext+1:2*iext , 1:iysize)
    expimage_ext(ixsize:ixsize-iext+1:-1 , 1:iysize) = &
         expimage_ext(ixsize-2*iext+1:ixsize-iext , 1:iysize)
    expimage_ext(1:ixsize , iext:1:-1) = &
         expimage_ext(1:ixsize , iext+1:2*iext)
    expimage_ext(1:ixsize , iysize:iysize-iext+1:-1) = &
         expimage_ext(1:ixsize , iysize-2*iext+1:iysize-iext)
    endif


!!! SMOOTH
    snrmin=db_snrmin
    sigma=db_smoothsigma

    allocate(mindex(ixsize,iysize),stat=status)
    allocate(kernel(ixsize,iysize),stat=status)
    allocate(image1(ixsize,iysize),stat=status)
    allocate(image2(ixsize,iysize),stat=status)
    allocate(image3(ixsize,iysize),stat=status)
    allocate(smoothed(ixsize,iysize),stat=status)
    allocate(msk_smoothed(ixsize,iysize),stat=status)
    allocate(exp_smoothed(ixsize,iysize),stat=status)
    allocate(snr(ixsize,iysize),stat=status)
    allocate(counts(ixsize,iysize),stat=status)
    mindex = 0
    kernel = 0.0
    image1 = 0.0
    image2 = 0.0
    image3 = 0.0
    smoothed = 0.0
    msk_smoothed = 0.0
    exp_smoothed = 0.0
    snr    = 0.0
    counts = 0.0

    allocate(bkgarr(nsig,ixsize,iysize))
    allocate(snrarr(nsig,ixsize,iysize))
    bkgarr = 0.0
    snrarr = 0.0

    if (status == 0) then

       image1 = image_ext * cheesemask_ext
       image2 = cheesemask_ext
       if (db_exp) then
          where (expimage_ext(:,:).ne.0) &
               image1(:,:) = image1(:,:)/expimage_ext(:,:)
          image3 = expimage_ext
       endif

       ! convolve with gaussian kernel of increasing width
       do isig=1,nsig

          call kgauss(sigma,ixsize,iysize,kernel,knorm)

          call  fconvolve(image1, image2, image3, kernel, &
                          smoothed, msk_smoothed, exp_smoothed, &
                          ixsize, iysize, db_exp)
          if (db_exp) then
             counts = smoothed / knorm * exp_smoothed
          else
             counts = smoothed / knorm
          endif
          snr = sqrt(counts)
          snrarr(isig,:,:) = snr

          where (msk_smoothed(:,:).ne.0) &
               bkgarr(isig,:,:) = smoothed / msk_smoothed
          sigma = sigma * sqrt2
       enddo

       ! Find SNR value nearest to snrmin:
       mindex = minloc(abs(snrarr-snrmin),dim=1)

       backmap_ext(:,:)=0.0
       do iy=1,iysize
          do ix=1,ixsize

             iminarr=minloc(abs(snrarr(:,ix,iy)-snrmin))
             imin=iminarr(1)
             snrdiff = snrarr(imin,ix,iy)-snrmin
             ! interpolation
             if (snrdiff > 0.0) then
                if (imin == 1  .or. abs(snrdiff) < snreps) then
                   backmap_ext(ix,iy) = bkgarr(imin,ix,iy)
                else
                   snrup=snrdiff
                   snrlow = abs(snrmin-snrarr(imin-1,ix,iy))
                   backmap_ext(ix,iy) = &
                        (snrlow*bkgarr(imin,ix,iy) &
                        + snrup*bkgarr(imin-1,ix,iy))/(snrup+snrlow)
                endif
             else
                if (imin == nsig .or. abs(snrdiff) < snreps) then
                   backmap_ext(ix,iy) = bkgarr(imin,ix,iy)
                else
                   snrlow=abs(snrdiff)
                   snrup = abs(snrarr(imin+1,ix,iy)-snrmin)
                   backmap_ext(ix,iy) = &
                        (snrup*bkgarr(imin,ix,iy) &
                        + snrlow*bkgarr(imin+1,ix,iy))/(snrup+snrlow)
                endif
             endif

          enddo
       enddo

    else
       call error("notEnoughMemory", &
            "Fit method smooth: Not enough memory to allocate array.")
    endif

    ! copy
    if (db_exp) then
       IBACK(ix0:ix1,iy0:iy1) = backmap_ext(ixmin:ixmax, iymin:iymax) &
                              * expimage_ext(ixmin:ixmax, iymin:iymax)
    else
       IBACK(ix0:ix1,iy0:iy1) = backmap_ext(ixmin:ixmax, iymin:iymax)
    endif

    ! catch NaNs which are not reset in restor.f by newer compiler versions
    ! (cf SPR-7587 Jan 2021) and throw a warning
    if (any(isnan(iback(ix0:ix1,iy0:iy1)))) &
         call warning("InvalidDataPoints", "Null values in the background map&
                       & found and reset to 1e-5. They could be caused by&
                       & bad bright pixels in the input data.")
    where(isnan(iback(ix0:ix1,iy0:iy1))) iback(ix0:ix1,iy0:iy1) = 1e-5

    ! Consistency check whether count number in input cheesed image and output
    ! background map differ too much. Limit currently: ten per cent
    ctest = abs(sum(iback*mfeld)/sum(iches*mfeld)-1.0)

    if (ctest > climit) call warning("InconsistentCountNumbers", &
         "Number of counts in background map deviates by more than ten per&
        & cent from the input image. E.g. check the image for bright pixels,&
        & the smoothsigma parameter, the exposure threshold of the input mask.")

  end subroutine esp_smooth



!-------------------------------------------------------------------------------
! KGAUSS
  subroutine kgauss(sigma,ixsize,iysize,kernel,norm)

    !---------------------------------------------
    ! calculate gaussian kernel
    ! rearrange quadrants for FFTW input:
    !    1|2           4|3
    !    ---    =>     ---
    !    3|4           2|1
    !---------------------------------------------


    real                        :: sigma
    real, dimension(:,:)        :: kernel
    integer                      :: ixsize
    integer                      :: iysize

    integer                      :: i,j,ii,jj,ii2,jj2,ncalc
    real                        :: x,y,twosig2,norm

    real  :: pi=3.141592654

    ncalc=min(int(5.*sigma),iysize/2)
    twosig2 = 2.0*sigma*sigma
    norm=1./(pi * 2.0*sigma*sigma)
    do j=1,ncalc
       y=real(j)-0.5
       jj=j
       jj2=iysize-j+1
       do  i=1,j
          x=real(i)-0.5
          ii=i
          ii2=ixsize-i+1
          kernel(ii,jj)   = norm*exp(-(x*x+y*y)/twosig2)
          kernel(jj,ii)   = kernel(ii,jj)
          kernel(ii2,jj)  = kernel(ii,jj)
          kernel(jj2,ii)  = kernel(ii,jj)
          kernel(ii2,jj2) = kernel(ii,jj)
          kernel(jj2,ii2) = kernel(ii,jj)
          kernel(ii,jj2)  = kernel(ii,jj)
          kernel(jj,ii2)  = kernel(ii,jj)
       enddo
    enddo

  end subroutine kgauss



!-------------------------------------------------------------------------------
! FCONVOLVE
  subroutine fconvolve(image1, image2, image3, kernel, &
                       smoothed1, smoothed2, smoothed3, &
                       ixsize, iysize, db_exp)

    real, dimension(:,:)   :: image1
    real, dimension(:,:)   :: image2
    real, dimension(:,:)   :: image3
    real, dimension(:,:)   :: smoothed1
    real, dimension(:,:)   :: smoothed2
    real, dimension(:,:)   :: smoothed3
    real, dimension(:,:)   :: kernel

    integer                :: ixsize,hxsize
    integer                :: iysize,hysize

    integer                :: status

    complex, dimension(:,:), allocatable       :: ffin, ffout, ffkernel
    complex, dimension(:,:), allocatable       :: ffmul,ffmskmul,ffexpmul
    complex, dimension(:,:), allocatable, save :: ffima,ffmsk,ffexp
    real,    dimension(:,:), allocatable       :: re_in,re_out

    integer*8,save         :: fplan ,rplan    !!! integer(kind=int64)
    real                   :: fftscale
    integer, parameter     :: fftw_measure=0
    integer, parameter     :: fftw_exhaustive=8
    integer, parameter     :: fftw_patient=32
    integer, parameter     :: fftw_estimate=64
    integer, parameter     :: fftw_estimate_patient=128

    integer                :: plan_type

    logical                :: db_exp
    logical,save           :: firstfc=.true.


    fftscale=1./float((ixsize*iysize))

    status=0

    hxsize=ixsize/2
    hysize=iysize/2

    if (.not. allocated(re_in))  allocate(re_in(ixsize,iysize),stat=status)
    if (.not. allocated(re_out)) allocate(re_out(ixsize,iysize),stat=status)
    if (.not. allocated(ffin))   allocate(ffin(1:hxsize+1,1:iysize),stat=status)
    if (.not. allocated(ffout))  allocate(ffout(1:hxsize+1,1:iysize),stat=status)

    if (.not. allocated(ffkernel)) allocate(ffkernel(1:hxsize+1,1:iysize),stat=status)
    if (.not. allocated(ffmul))    allocate(ffmul(1:hxsize+1,1:iysize),stat=status)
    if (.not. allocated(ffmskmul)) allocate(ffmskmul(1:hxsize+1,1:iysize),stat=status)
    if (.not. allocated(ffexpmul)) allocate(ffexpmul(1:hxsize+1,1:iysize),stat=status)

    if (firstfc) then
       allocate(ffima(1:hxsize+1,1:iysize),stat=status)
       allocate(ffmsk(1:hxsize+1,1:iysize),stat=status)
       allocate(ffexp(1:hxsize+1,1:iysize),stat=status)
    endif

    if (status == 0) then

       ! note: FFTW_MEASURE modifies input array
       plan_type=FFTW_ESTIMATE

       call sfftw_plan_dft_r2c_2d(rplan,ixsize,iysize,re_in,ffout,plan_type)
       call sfftw_plan_dft_c2r_2d(fplan,ixsize,iysize,ffin,re_out,plan_type)

       if (firstfc) then
          re_in=image1
          call sfftw_execute(rplan)
          ffima=ffout

          re_in=image2
          call sfftw_execute(rplan)
          ffmsk=ffout

          if (db_exp) then
          re_in=image3
          call sfftw_execute(rplan)
          ffexp=ffout
          endif
       endif

       re_in=kernel
       call sfftw_execute(rplan)
       ffkernel=ffout

       ffmul    = ffima*ffkernel*fftscale
       ffmskmul = ffmsk*ffkernel*fftscale
       if (db_exp) then
       ffexpmul = ffexp*ffkernel*fftscale
       endif

       ffin = ffmul
       call sfftw_execute(fplan)
       smoothed1 = re_out

       ffin = ffmskmul
       call sfftw_execute(fplan)
       smoothed2 = re_out

       if (db_exp) then
       ffin = ffexpmul
       call sfftw_execute(fplan)
       smoothed3 = re_out
       endif

       firstfc=.false.

    else
       call error("notEnoughMemory", &
            "Fit method smooth: Not enough memory to allocate array.")
    endif

  end subroutine fconvolve




end module esp_smooth_mod
